import arxiv
from tqdm import tqdm
from driarxiv.parser import PDFWorker
from openai import OpenAI
from oaib import Auto
from dria import Dria, Models
from driarxiv.logger_config import logger

import os


class OpenAIWorker:
    def __init__(self):
        self.client = OpenAI()
        self.batch = Auto(workers=8, silent=True)

    async def get_titles(self, references):
        for ref in references:
            prompt = 'Return the paper title from given passage. ' \
                     'Only prompt the title without any other info including writer names. Passage:{}'.format(
                ref)
            await self.batch.add(
                "chat.completions.create",
                model="gpt-3.5-turbo",
                messages=[{"role": "user", "content": prompt}]
            )
        results = await self.batch.run()
        results = results.get("result")
        titles = [response["choices"][0]["message"]["content"].replace('"',"").strip() for response in results]
        return titles


class ArxivWorker:
    def __init__(self, tag):
        self.client = arxiv.Client()
        self.dir_path = tag + "/references"

    def _get_downloaded_files(self):
        return os.listdir(self.dir_path)

    def download_file(self, paper):
        files = self._get_downloaded_files()
        fname = "{}.pdf".format(paper.title)
        if fname not in files:
            paper.download_pdf(filename=fname, dirpath=self.dir_path)

    def get_paper_from_url(self, url: str, download: bool = False):
        id = url.split("/")[-1].replace(".pdf", "")
        if "." not in id:
            id = url.split("/")[-2] + "/" + id
        search = arxiv.Search(id_list=[id], max_results=1)
        results = self.client.results(search)
        papers = list(results)
        paper = papers[0]
        if download:
            self.download_file(paper)
        return paper

    def search_for_paper(self, title: str, get_first: bool = True):
        search = arxiv.Search(
            query='ti:"{}"'.format(title),
            max_results=10,
            sort_by=arxiv.SortCriterion.SubmittedDate
        )
        results = self.client.results(search)
        papers = list(results)
        if len(papers) == 0:
            return None
        if get_first:
            return [papers[0]]
        return papers


class DriaArxiv:
    def __init__(self, dria_api_key, tag:str):
        self.pdf_worker = PDFWorker()
        self.tag = tag
        self.opeanai_worker = OpenAIWorker()
        self.arxiv_worker = ArxivWorker(tag)
        self.dria_index = Dria(api_key=dria_api_key)

    @staticmethod
    def filename(paper):
        return "{}.pdf".format(paper.title)

    def get_chunks_and_references(self, fname):
        return self.pdf_worker.from_file(fname, self.tag)

    async def create_files_and_chunks(self, paper, depth=1):
        if depth > 3:
            raise ValueError("Depth cannot be more than 3")

        logger.info("Creating files and chunks for paper: {}".format(paper.title))
        data = []
        worked_on = []
        original_chunks, original_references = self.get_chunks_and_references(self.filename(paper))

        paper_titles = await self.opeanai_worker.get_titles(original_references)

        logger.info("Adding origin paper's chunks: %s\n", paper.title)

        for chunk in tqdm(original_chunks):
            data.append({"text": chunk, "metadata": {"text": chunk, "title": paper.title}})

        worked_on.append(paper.title)

        d = 0
        while d < depth:
            new_refs = []
            logger.info("Working on references...")
            for title in tqdm(paper_titles):
                _p = self.arxiv_worker.search_for_paper(title)
                if _p is None:
                    continue
                _p = _p[0]

                try:
                    self.arxiv_worker.download_file(_p)
                except:
                    try:
                        self.arxiv_worker.download_file(_p)
                    except:
                        continue

                if _p.title in worked_on:
                    continue
                else:
                    chunks_, refs_ = self.get_chunks_and_references(self.filename(_p))

                    for chunk in chunks_:
                        data.append({"text": chunk, "metadata": {"text": chunk, "title":_p.title}})

                    worked_on.append(_p.title)

                    if d+1 < depth:
                        paper_titles = await self.opeanai_worker.get_titles(refs_)
                        new_refs += paper_titles

            paper_titles = list(set(new_refs))
            d += 1

        return data

    async def create_wiki(self, url):

        def chunked_list(data, n):
            for i in range(0, len(data), n):
                yield data[i:i + n]

        paper = self.arxiv_worker.get_paper_from_url(url, download=True)
        data = await self.create_files_and_chunks(paper)

        contract_id = self.dria_index.create(
            name=paper.title,
            embedding=Models.bge_base_en,
            category=paper.primary_category,
            description=paper.summary)

        for batch in tqdm(chunked_list(data, 1000), desc="Inserting chunks into index"):
            _ = self.dria_index.insert_text(
                batch
            )

        return contract_id



import re
from abc import ABC, abstractmethod
from io import BytesIO

import json


class FileWorker(ABC):

    @abstractmethod
    def to_chunks(self, fname):
        pass

    @staticmethod
    def chunk_list(lst, chunk_size):
        """Yield successive chunk_size chunks from lst."""
        for i in range(0, len(lst), chunk_size):
            yield lst[i:i + chunk_size]

    @staticmethod
    def read_response_bytes(response):
        stream = BytesIO()
        for chunk in response.iter_content(chunk_size=1024 * 1024):  # Process the stream in chunks of 1MB
            stream.write(chunk)
        stream.seek(0)

        return stream

    @staticmethod
    def remove_references(text):
        import re
        cleaned = re.sub(r'\[\d+\]', '', text)
        cleaned = ' '.join(cleaned.split())
        return cleaned

    def clean_text(self, text):
        if text == "":
            return text
        text = text.replace('- ', '').replace('\n', ' ')
        text = text.replace('/*', '').replace('*/', '').replace('@', '')
        text = text.replace('w/', 'with ')
        text = ' '.join(text.split())
        text = self.remove_references(text)
        text = self.remove_urls(text)
        if text[0].islower():
            try:
                text = self.split_after_first_char_regex(text, ['.', '?', '!', "'"])
            except Exception as e:
                return text
        return text

    @staticmethod
    def remove_urls(text):
        return re.sub(r'http[s]?://\S+', '', text)

    @staticmethod
    def split_after_first_char_regex(text, char_list):
        pattern = '[' + re.escape(''.join(char_list)) + ']'
        match = re.search(pattern, text)

        if match:
            position = match.end()
            return text[position:]
        else:
            return ''





import argparse
import os

from driarxiv.worker import DriaArxiv
from driarxiv.build_readme import run
import asyncio
import logging
from driarxiv.logger_config import logger


def main():

    logging.getLogger('asyncio').setLevel(
        logging.WARNING)  # Remove asyncio debug and info messages, but leave warnings.
    logging.getLogger("requests").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("httpx").setLevel(logging.WARNING)
    logging.getLogger("arxiv").setLevel(logging.WARNING)

    parser = argparse.ArgumentParser(description='Driarxiv a CLI to generate a comprehensive wiki from arXiv papers.')
    subparsers = parser.add_subparsers(dest='command')

    # Add command
    parser_add = subparsers.add_parser('generate', help="Create a docs from paper and it's references.")
    parser_add.add_argument('arxiv_url', type=str, help='URL of the paper')

    # Add command
    parser_add = subparsers.add_parser('from_rag', help="Create a docs from paper and it's references.")
    parser_add.add_argument('arxiv_url', type=str, help='URL of the paper')
    parser_add.add_argument('contract_id', type=str, help='contract_id of dria index')

    args = parser.parse_args()

    if args.command == 'generate':
        url = args.arxiv_url
        tag = url.split("/")[-1].replace(".pdf", "")
        logger.info("Working on %s", tag)
        # Create docs directory if it does not exist
        if not os.path.exists(tag):
            os.makedirs(tag)
        if not os.path.exists(tag + "/docs"):
            os.makedirs(tag + "/docs")

        if not os.path.exists(tag + "/references"):
            os.makedirs(tag + "/references")

        da = DriaArxiv(dria_api_key=os.environ["DRIA_API_KEY"], tag=tag)
        contract_id = asyncio.run(da.create_wiki(url))
        run(url, contract_id, tag)

    elif args.command == 'from_rag':
        url = args.arxiv_url
        tag = url.split("/")[-1].replace(".pdf", "")

        if not os.path.exists(tag):
            os.makedirs(tag)
        if not os.path.exists(tag + "/docs"):
            os.makedirs(tag + "/docs")

        if not os.path.exists(tag + "/references"):
            os.makedirs(tag + "/references")

        url = args.arxiv_url
        contract_id = args.contract_id
        run(url, contract_id, tag)

if __name__ == '__main__':
    main()



import logging


# Configure logging
logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                    datefmt='%m/%d/%Y %I:%M:%S %p',
                    handlers=[
                        logging.FileHandler("driarxiv.log"),
                        logging.StreamHandler()
                    ])

logger = logging.getLogger(__name__)


from abc import ABC

from driarxiv.parser.base import FileWorker
import time
from tqdm import tqdm
import requests

class PDFWorker(FileWorker):

    def to_chunks(self, fname):
        from unstructured.partition.pdf import partition_pdf
        from unstructured.chunking.title import chunk_by_title
        import re
        from unstructured.cleaners.core import clean_non_ascii_chars, clean_extra_whitespace, group_broken_paragraphs, \
            replace_unicode_quotes
        para_split_re = re.compile(r"(\s*\n\s*){3}")
        response = requests.get(fname)
        response_bytes = self.read_response_bytes(response)

        elements = partition_pdf(file=response_bytes, include_page_breaks=False, strategy="fast")
        for el in elements:
            el.apply(replace_unicode_quotes)
            el.apply(clean_extra_whitespace)
            el.apply(clean_non_ascii_chars)

        ind = 0
        for i, el in enumerate(elements):
            if "REFERENCES" in el.text:
                ind = i
                break

        chunks = chunk_by_title(elements[:ind], max_characters=1000)

        for chunk in chunks:
            chunk.text = group_broken_paragraphs(chunk.text, paragraph_split=para_split_re)
            chunk.text = self.clean_text(chunk.text)

        references = elements[ind:]
        references = [ref.text for ref in references if len(ref.text) > 20]

        return [chunk.text for chunk in chunks], references

    def from_file(self, fname, tag):
        from unstructured.partition.pdf import partition_pdf
        from unstructured.chunking.title import chunk_by_title
        import re
        from unstructured.cleaners.core import clean_non_ascii_chars, clean_extra_whitespace, group_broken_paragraphs, \
            replace_unicode_quotes
        para_split_re = re.compile(r"(\s*\n\s*){3}")

        elements = partition_pdf(filename= tag + "/references/" + fname, include_page_breaks=False, strategy="fast")
        for el in elements:
            el.apply(replace_unicode_quotes)
            el.apply(clean_extra_whitespace)
            el.apply(clean_non_ascii_chars)

        ind = 0
        for i, el in enumerate(elements):
            if "REFERENCES" in el.text or "References" in el.text:
                ind = i
                break

        chunks = chunk_by_title(elements[:ind], max_characters=1000)

        for chunk in chunks:
            chunk.text = group_broken_paragraphs(chunk.text, paragraph_split=para_split_re)
            chunk.text = self.clean_text(chunk.text)

        references = elements[ind:]
        references = [ref.text for ref in references if len(ref.text) > 20]

        return [chunk.text for chunk in chunks if len(chunk.text)>10], references



from setuptools import setup, find_packages

setup(
    name='driarxiv',
    version='1.0.2',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'driarxiv=driarxiv.run:main',
        ],
    },
    install_requires=[
        'openai>=0.10.2',
        'arxiv>=1.3.0',
        'tqdm>=4.61.2',
        'dria>=0.1.13',
        'unstructured',
        'unstructured[pdf]',
        'oaib',
        'tiktoken'
    ],
    python_requires='>=3.9',
    description='driarxiv: A CLI tool to generate a comprehensive wiki from arXiv papers.',
    author='andthattoo',
    author_email='omer@firstbatch.xyz',
    url='https://github.com/andthattoo/driarxiv',
)


from dria import Dria
from openai import OpenAI
from driarxiv.worker import ArxivWorker
from driarxiv.parser import PDFWorker
import re
import tiktoken
from time import time
from tqdm import tqdm
import os
from driarxiv.logger_config import logger


def filename_from_title(title):
    return title.lower().replace(" ", "_").replace(":", "").\
        replace("?", "").replace("!", "").replace("(", "").\
        replace(")", "").replace("/", "").replace("\\", "").\
        replace("'", "").replace('"', "").replace("*", "").replace(".", "")

def replace_term_with_link(text, term, fname):
    escaped_term = re.escape(term)
    replacement = r'[\g<0>](docs/{}.md)'.format(fname.lower())
    result = re.sub(r'\b' + escaped_term + r'\b', replacement, text)

    return result


def replace_term_with_link_in_docs(text, term, fname):
    escaped_term = re.escape(term)
    replacement = r'[\g<0>]({}.md)'.format(fname.lower())
    result = re.sub(r'\b' + escaped_term + r'\b', replacement, text)

    return result


def get_context(question, dria_client, encoder):
    ctx = dria_client.search(question, top_n=20, rerank=True, level=0)
    ctx = " ".join([c["metadata"] for c in ctx if c["score"] > 0.8])
    ctx = encoder.decode(encoder.encode(ctx)[:8130])
    return ctx


class OpenAIWorker:
    def __init__(self):
        self.client = OpenAI()
        self.messages = [{"role": "system", "content": "You are a helpful assistant."}]

    def hard_reset(self):
        self.messages = [{"role": "system", "content": "You are a helpful assistant."}]

    def ask_gpt(self, question, context):
        p = """Use the following pieces of context to answer the user question. This context retrieved from a knowledge base and you should use only the facts from the context to answer.
        Your answer must be based on the context. If the context not contain the answer, just say that 'I don't know', don't try to make up an answer, use the context.
        Don't address the context directly, but use it to answer the user question like it's your own knowledge.
        
        While answering, use technical language and be concise. Add '$' around latex math equations.

        Context:
        """
        p += context
        p += "\n"
        p += "Question: {}".format(question)

        self.messages.append({"role": "user", "content": p})
        response = self.client.chat.completions.create(
            model="gpt-4-0125-preview",
            messages=self.messages)

        self.messages.append({"role": "system", "content": response.choices[0].message.content})

        return response.choices[0].message.content

    def ner(self, readme):

        response = self.client.chat.completions.create(
            model="gpt-4-0125-preview",
            messages=[
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "Return the named entities from given passage the way they are written inside the text."
                                            "Directly write them without letters, number or bulletpoints, write them line by line. "
                                            "If two entities are same but written differently (like abbravations) write only one of them." 
                                            "For instance write 'United States' and 'USA' as 'United States' instead of two lines."
                                            "For instance write 'John Coltrane's and 'John Coltrane' as 'John Coltrane' instead of two lines."
                                            "For instance write 'Large Language Model' and 'LLMs' as 'Large Language Model' instead of two lines."
                                            "Seperate them by '\n' Passage:{}".format(readme)
                 },
            ]
        )
        self.messages.append({"role": "system", "content": response.choices[0].message.content})

        return response.choices[0].message.content


def run(url, contract_id, tag):
    st = time()
    # Create dir
    encoder = tiktoken.encoding_for_model("gpt-4")
    da = Dria(api_key=os.environ["DRIA_API_KEY"])
    da.set_contract(contract_id)
    arxiv_worker = ArxivWorker(tag)
    pdfworker = PDFWorker()
    oaw = OpenAIWorker()

    readmes = {}
    readmes["main"] = []

    paper = arxiv_worker.get_paper_from_url(url, download=True)
    fname = "{}.pdf".format(paper.title)

    # First ask what is the paper about
    logger.info("Asking gpt what the paper is about.")
    question = "What is {}? Write a comprehensive answer".format(paper.title)
    context = get_context(question, da, encoder)
    response = oaw.ask_gpt(question, context)
    readmes["main"].append(response)

    # Ask about the methodology outline
    logger.info("Asking gpt about the methodologies of the paper.")
    question = "List the methodology of {}, outline processes and explain step by step. Use technical language, but make it concise".format(paper.title)
    sub_q = "What is the methodology of {}?".format(paper.title)
    context = get_context(sub_q, da, encoder)
    response = oaw.ask_gpt(question, context)
    readmes["main"].append(response)

    logger.info("Asking gpt about the results and benchmarks")
    question = "What are the results and benchmarks of {}?. Make it concise".format(paper.title)
    context = da.search(question, top_n=20, rerank=False, level=0)
    context = " ".join([c["metadata"]["text"] for c in context if c["score"] > 0.65 and c["metadata"]["title"] == paper.title])
    context = encoder.decode(encoder.encode(context)[:8191 - len(encoder.encode(question))])
    response = oaw.ask_gpt(question, context)
    readmes["main"].append(response)
    oaw.hard_reset()

    # Ask gpt to create questions about paper
    logger.info("Asking gpt to create fundamental questions about the paper")
    question = "Create a list of fundamental questions you think needs to be explored for given context to be understood. No more than 3 questions. Only ask questions if you think the are neccessary for understanding the underyling mechanisms. If the question you asked is already answered within context, dont ask it. Seperate questions by a new line."
    context = " ".join(readmes["main"])
    context = encoder.decode(encoder.encode(context)[:8130])
    response = oaw.ask_gpt(question, context)
    qs = response.split("\n")
    qs = [q for q in qs if len(q) > 7]

    logger.info("GPT answering the questions")
    for question in qs:
        context = get_context(question, da, encoder)
        response = oaw.ask_gpt(question, context)
        readmes["main"].append(question)
        readmes["main"].append(response)
        oaw.hard_reset()


    master_readme = "\n".join(readmes["main"])
    named_entities = oaw.ner(master_readme)
    named_entities = list(set(named_entities.split("\n")))
    logger.info("Creating readmes for named entities")
    for entity in tqdm(named_entities):
        question = "How {} is related to {}? Explain profoundly, with technical details".format(entity, paper.title)
        context = get_context(entity, da, encoder)
        if context == "":
            continue
        readmes[entity] = []
        response = oaw.ask_gpt(question, context)
        readmes[entity].append(response)
        oaw.hard_reset()

    logger.info("Writing readmes to disk")
    all_readmes = {k: "\n".join(v) for k,v in readmes.items() if k != "main"}
    all_readmes = {k:v for k,v in all_readmes.items() if "I don't know" not in v}

    for k, v in all_readmes.items():

        ks = set(all_readmes.keys())
        ks.remove(k)
        for key in ks:
            fname = filename_from_title(key)#key.lower().replace(" ", "_")
            v = replace_term_with_link_in_docs(v, key, fname)

        fname = filename_from_title(k) #k.lower().replace(" ", "_")
        with open(tag + "/docs/{}.md".format(fname), "w") as f:
            f.write("{}".format(v))
        master_readme = replace_term_with_link(master_readme, k, fname)

    with open(tag + "/README.md".format(tag), "w") as f:
        f.write(master_readme)

    logger.info("Done in {} seconds".format(time()-st))



if __name__ == "__main__":
    run("https://arxiv.org/pdf/2310.11454.pdf", "gl5XQxrYWrqS4xgak24W1vy8ORGilOBHPU_K2CjVfmo")